# coding=utf-8
# Copyright 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.


"""
Generate C source code from Khronos XML.
"""

import argparse
import os.path
import re
import sys
import functools
from collections import namedtuple

import mako.runtime
import mako.template

PIGLIT_TOP_DIR = os.path.join(os.path.dirname(__file__), '..', '..')
sys.path.append(PIGLIT_TOP_DIR)

import registry.gl  # pylint: disable=import-error


debug = False


def log_debug(msg):
    if debug:
        prog_name = os.path.basename(sys.argv[0])
        print('debug: {0}: {1}'.format(prog_name, msg), file=sys.stderr)


def main():
    global debug

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-o', '--out-dir', required=True)
    argparser.add_argument('-d', '--debug', action='store_true', default=False)
    args = argparser.parse_args()

    debug = args.debug
    registry.gl.debug = debug

    gl_registry = registry.gl.parse()
    DispatchCode.emit(args.out_dir, gl_registry)
    EnumCode.emit(args.out_dir, gl_registry)


class DispatchCode(object):

    H_TEMPLATE = 'piglit-dispatch-gen.h.mako'
    C_TEMPLATE = 'piglit-dispatch-gen.c.mako'

    Api = namedtuple('DispatchApi',
                     ('name', 'base_version_int', 'c_piglit_token'))

    APIS = {
        'gl': Api('gl', 10, 'PIGLIT_DISPATCH_GL'),
        'gles1': Api('gles1', 11, 'PIGLIT_DISPATCH_ES1'),
        'gles2': Api('gles2', 20, 'PIGLIT_DISPATCH_ES2'),
    }
    APIS['glcore'] = APIS['gl']

    assert set(APIS.keys()) | set(['glsc2']) | set(['disabled']) == set(registry.gl.VALID_APIS)

    @classmethod
    def emit(cls, out_dir, gl_registry):
        assert isinstance(gl_registry, registry.gl.Registry)
        context_vars = dict(dispatch=cls, gl_registry=gl_registry)
        render_template(cls.H_TEMPLATE, out_dir, **context_vars)
        render_template(cls.C_TEMPLATE, out_dir, **context_vars)


def render_template(filename, out_dir, **context_vars):
    assert filename.endswith('.mako')
    template_filepath = os.path.join(os.path.dirname(__file__), filename)
    out_filepath = os.path.join(out_dir, os.path.splitext(filename)[0])

    warning = 'DO NOT EDIT! Script {0!r} generated this file from {1!r}'
    warning = warning.format(os.path.basename(__file__), filename)

    fake_alignment = re.compile(r'\.*\n\.+', flags=re.MULTILINE)
    fake_tab = re.compile(r'>-------')

    def fake_whitespace(proto_text):
        if debug:
            print('fake whitespace: before: {0!r}'.format(proto_text))

        text = proto_text
        text = fake_alignment.sub('', text)
        text = fake_tab.sub('\t', text)
        if debug:
            print('fake whitespace:  after: {0!r}'.format(text))
        return text

    with open(out_filepath, 'w') as out_file:
        template = mako.template.Template(
            filename=template_filepath,
            strict_undefined=True)
        ctx = mako.runtime.Context(
            buffer=out_file,
            warning=warning,
            fake_whitespace=fake_whitespace,
            **context_vars)
        template.render_context(ctx)


class EnumCode(object):

    C_TEMPLATE = 'piglit-util-gl-enum-gen.c.mako'

    @classmethod
    def emit(cls, out_dir, gl_registry):
        assert isinstance(gl_registry, registry.gl.Registry)
        enums = cls.get_enums_in_default_namespace(gl_registry)
        unique_enums = cls.get_unique_enums(enums)
        enums_by_name = cls.get_enums_by_name(enums)
        memory_barrier = cls.get_enums_in_memory_barrier_group(gl_registry)
        render_template(
            cls.C_TEMPLATE,
            out_dir,
            gl_registry=gl_registry,
            sorted_unique_enums_in_default_namespace=unique_enums,
            sorted_enums_by_name=enums_by_name,
            sorted_by_name_memory_barrier_enums=memory_barrier)

    @classmethod
    def get_enums_in_default_namespace(cls, gl_registry):
        enums = []
        for enum_group in gl_registry.enum_groups:
            if enum_group.type == 'default_namespace':
                for enum in enum_group.enums:
                    enums.append(enum)
        return enums

    @classmethod
    def get_enums_in_memory_barrier_group(cls, gl_registry):
        enums = []
        for enum_group in gl_registry.enum_groups:
            if enum_group.name == 'MemoryBarrierMask':
                if enum_group.type == 'bitmask':
                    for enum in enum_group.enums:
                        enums.append(enum)
        return cls.get_enums_by_name(enums)

    @classmethod
    def get_unique_enums(cls, enums):
        def append_enum_if_new_value(enum_list, enum):
            if enum_list[-1].num_value < enum.num_value:
                enum_list.append(enum)
            return enum_list

        enums = sorted(enums)
        enums = functools.reduce(append_enum_if_new_value, enums[1:], [enums[0]])
        return enums

    @classmethod
    def get_enums_by_name(cls, enums):
        enums = sorted(enums, key=lambda enum: enum.name)
        return enums


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2020, 2022 Collabora Ltd
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import base64
import hashlib
import hmac
import xml.etree.ElementTree as ET
from email.utils import formatdate
from os import path, remove
from pathlib import Path
from typing import Any, Dict, List

import requests
from requests.adapters import HTTPAdapter, Retry
from requests.utils import requote_uri

from framework import core, exceptions
from framework.replay.local_file_adapter import LocalFileAdapter
from framework.replay.options import OPTIONS

__all__ = ['ensure_file']

minio_credentials = None


def sign_with_hmac(key, message):
    key = key.encode("UTF-8")
    message = message.encode("UTF-8")

    signature = hmac.new(key, message, hashlib.sha1).digest()

    return base64.encodebytes(signature).strip().decode()


def get_minio_credentials(url):
    global minio_credentials

    if minio_credentials is not None:
        return (minio_credentials['AccessKeyId'],
                minio_credentials['SecretAccessKey'],
                minio_credentials['SessionToken'])

    minio_credentials = {}

    params = {'Action': 'AssumeRoleWithWebIdentity',
              'Version': '2011-06-15',
              'RoleArn': 'arn:aws:iam::123456789012:role/FederatedWebIdentityRole',
              'RoleSessionName': OPTIONS.download['role_session_name'],
              'DurationSeconds': 3600,
              'WebIdentityToken': OPTIONS.download['jwt']}
    r = requests.post('https://%s' % OPTIONS.download['minio_host'], params=params)
    if r.status_code >= 400:
        print(r.text)
    r.raise_for_status()

    root = ET.fromstring(r.text)
    for attr in root.iter():
        if attr.tag == '{https://sts.amazonaws.com/doc/2011-06-15/}AccessKeyId':
            minio_credentials['AccessKeyId'] = attr.text
        elif attr.tag == '{https://sts.amazonaws.com/doc/2011-06-15/}SecretAccessKey':
            minio_credentials['SecretAccessKey'] = attr.text
        elif attr.tag == '{https://sts.amazonaws.com/doc/2011-06-15/}SessionToken':
            minio_credentials['SessionToken'] = attr.text

    return (minio_credentials['AccessKeyId'],
            minio_credentials['SecretAccessKey'],
            minio_credentials['SessionToken'])


def get_minio_authorization_headers(url, resource):
    minio_key, minio_secret, minio_token = get_minio_credentials(url)

    date = formatdate(timeval=None, localtime=False, usegmt=True)
    to_sign = "GET\n\n\n%s\nx-amz-security-token:%s\n/%s/%s" % (date,
                                                                minio_token,
                                                                OPTIONS.download['minio_bucket'],
                                                                requote_uri(resource))
    signature = sign_with_hmac(minio_secret, to_sign)

    headers = {'Host': OPTIONS.download['minio_host'],
               'Date': date,
               'Authorization': 'AWS %s:%s' % (minio_key, signature),
               'x-amz-security-token': minio_token}
    return headers


def get_jwt_authorization_headers(url, resource):
    date = formatdate(timeval=None, localtime=False, usegmt=True)
    jwt = OPTIONS.download['jwt']

    headers = {'Date': date,
               'Authorization': 'Bearer %s' % (jwt)}
    return headers


def calc_etags(inputfile: Path, partsize: int = 5 * 1024 * 1024) -> List[str]:
    '''Calculate e-tag generated by FDO upload script (s3cp).'''
    md5 = hashlib.md5()
    md5_digests = []
    with open(inputfile, 'rb') as file:
        for chunk in iter(lambda: file.read(partsize), b''):
            md5.update(chunk)
            md5_digests.append(hashlib.md5(chunk).digest())
        return [
                hashlib.md5(b''.join(md5_digests)).hexdigest() + '-' + str(len(md5_digests)),
                md5.hexdigest()
                ]


def chase_redirects(session: requests.Session, request: requests.PreparedRequest) -> requests.Response:
    """Finds an HTTP response for a request, with custom redirect chasing

    When using a passthrough cache, requests does not see that the hostname
    in the URI has changed and does not strip the Authorization header. This
    breaks with fd.o's s3-proxy, which returns a redirect with pre-signed
    parameters in the request params. Without a proxy, python-requests
    detects that we have redirected to a different host and so strips the
    Authorization header, but with the proxy we do not strip the
    Authorization header, which causes Hetzner's object-storage backend to
    throw a 400.

    :param session: python-requests session object
    :param url: URL to download
    :param headers: Headers to send with the request
    """

    response = session.send(request, allow_redirects=False)
    if response is None or not response.is_redirect:
        return response
    assert response.next
    if "Authorization" in response.next.headers:
        del response.next.headers["Authorization"]
    return chase_redirects(session, response.next)


@core.timer_ms
def download(url: str, file_path: str, headers: Dict[str, str], attempts: int = 2) -> None:
    """Downloads a URL content into a file

    :param url: URL to download
    :param file_path: Local file name to contain the data downloaded
    :param attempts: Number of attempts
    """
    retries = Retry(
        backoff_factor=20,
        connect=attempts,
        read=attempts,
        redirect=0,
        status_forcelist=[429, 500, 502, 503, 504],
        raise_on_redirect=False
    )
    session = requests.Session()
    for protocol in ["http://", "https://"]:
        adapter = HTTPAdapter(max_retries=retries)
        session.mount(protocol, adapter)
    for protocol in ["file://"]:
        file_adapter = LocalFileAdapter()
        session.mount(protocol, file_adapter)

    md5 = hashlib.md5()
    local_file_checksums: List[Any] = []
    md5_digests = []
    request = session.prepare_request(requests.Request(method='GET', url=url, headers=headers))
    with chase_redirects(session, request) as response:
        response.raise_for_status()
        with open(file_path, "wb") as file:
            # chuck_size must be equal to s3cp upload chunk for md5 digest to match
            for chunk in response.iter_content(chunk_size=5 * 1024 * 1024):
                if chunk:
                    file.write(chunk)
                    md5.update(chunk)
                    md5_digests.append(hashlib.md5(chunk).digest())
            local_file_checksums = [
                    hashlib.md5(b''.join(md5_digests)).hexdigest() + '-' + str(len(md5_digests)),
                    md5.hexdigest()
            ]

    verify_file_integrity(file_path, response.headers, local_file_checksums)


def verify_file_integrity(file_path: str, headers: Any, local_file_checksums: Any) -> None:
    """
    :param file_path: path to the local file
    :param headers: reference to the request
    :param local_file_checksums: list of already generated MD5
    """
    try:
        remote_file_checksum: str = headers["etag"].strip('\"').lower()
        if remote_file_checksum not in local_file_checksums:
            remove(file_path)
            raise exceptions.PiglitFatalError(
                    f"MD5 checksum {local_file_checksums} "
                    f"doesn't match remote ETag MD5 {remote_file_checksum}, removing file..."
            )
    except KeyError:
        print("ETag is missing from the HTTPS header. "
              "Fall back to Content-length verification.")

    try:
        remote_file_size = int(headers["Content-Length"])
    except KeyError:
        print("Error getting Content-Length from server. "
              "Skipping file size check.")
        return

    local_file_size = path.getsize(file_path)
    if remote_file_size != local_file_size:
        remove(file_path)
        raise exceptions.PiglitFatalError(
                f"Invalid filesize src {remote_file_size} "
                f"doesn't match {local_file_size}, removing file..."
        )


def verify_local_file_checksum(url, file_path, headers, destination_file_path):
    @core.timer_ms
    def check_md5():
        print(
            f"[check_image] Verifying already downloaded file {file_path}",
            end=" ",
            flush=True,
        )
        verify_file_integrity(
            destination_file_path, remote_headers, calc_etags(destination_file_path)
        )

    print(f"[check_image] Requesting headers for {file_path}", end=" ", flush=True)
    try:
        response = requests.head(url + file_path, timeout=60, headers=headers)
    except requests.exceptions.RequestException as err:
        print(f"Not verified! HTTP request failed with {err}", flush=True)
        return
    print(
        f"returned {response.status_code}.",
        f"Took {response.elapsed.microseconds / 1000} ms",
        flush=True,
    )
    remote_headers = response.headers

    check_md5()


def ensure_file(file_path):
    destination_file_path = path.join(OPTIONS.db_path, file_path)
    if OPTIONS.download['url'] is None:
        if not path.exists(destination_file_path):
            raise exceptions.PiglitFatalError(
                '{} missing'.format(destination_file_path))
        return

    url = OPTIONS.download['url'].geturl()

    if OPTIONS.download['caching_proxy_url'] is not None:
        url = OPTIONS.download['caching_proxy_url'].geturl() + url

    core.check_dir(path.dirname(destination_file_path))

    if OPTIONS.download['minio_host']:
        assert OPTIONS.download['minio_bucket']
        assert OPTIONS.download['role_session_name']
        assert OPTIONS.download['jwt']
        headers = get_minio_authorization_headers(url, file_path)
    elif OPTIONS.download['jwt']:
        headers = get_jwt_authorization_headers(url, file_path)
    else:
        headers = None

    if not OPTIONS.download['force'] and path.exists(destination_file_path):
        verify_local_file_checksum(url, file_path, headers, destination_file_path)
        return

    print(f"[check_image] Downloading file {file_path}", end=" ", flush=True)

    download(url + file_path, destination_file_path, headers)



# coding=utf-8
# Copyright 2016, 2019 Intel Corporation
# Copyright 2013, 2014 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Authors: Tom Stellard <thomas.stellard@amd.com>
#

import re

from .base import Test

__all__ = [
    'GTest',
]


class GTest(Test):
    def interpret_result(self):
        # Since gtests can have several subtets, if any of the subtests fail
        # then we need to report fail.
        out = self.result.out
        if len(re.findall('FAILED', out, re.MULTILINE)) > 0:
            self.result.result = 'fail'
        elif len(re.findall('PASSED', out, re.MULTILINE)) > 0:
            self.result.result = 'pass'
        else:
            #If we get here, then the test probably exited early.
            self.result.result = 'fail'

        super(GTest, self).interpret_result()



# coding=utf-8
#
# Copyright (c) 2014, 2016-2017, 2019-2020 Intel Corporation
# Copyright (c) 2019 Collabora Ltd
# Copyright © 2019-2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT


""" Module providing a GFXReconstruct dump backend for replayer """

import os
import re
import subprocess
import sys
from os import path

from packaging import version

from framework import core, exceptions

from . import DumpBackendError
from .abstract import DumpBackend, dump_handler
from .register import Registry

__all__ = [
    'REGISTRY',
    'GFXReconstructBackend',
]


_MIN_VERSION = version.Version('0.9.4')
_VERSION_RE = re.compile(r'\s*GFXReconstruct Version\s*([0-9]\.[0-9]\.[0-9])')

_TOTAL_FRAMES_RE = re.compile(r'\s*Total frames:\s*([0-9]*)')


class GFXReconstructBackend(DumpBackend):
    """ replayer's GFXReconstruct dump backend

    This backend uses GFXReconstruct for replaying its traces.

    The path to the GFXReconstruct binary is configurable.

    It also admits configuration for passing extra parameters to
    GFXReconstruct.

    """

    def __init__(self, trace_path, output_dir=None, calls=None, **kwargs):
        super(GFXReconstructBackend, self).__init__(trace_path, output_dir,
                                                    calls, **kwargs)
        extension = path.splitext(self._trace_path)[1]

        if extension != '.gfxr':
            raise exceptions.PiglitFatalError(
                'Invalid trace_path: "{}" tried to be dumped '
                'by the GFXReconstructBackend.\n'.format(self._trace_path))

    def _get_last_frame_call(self):
        gfxrecon_info_bin = core.get_option(
            'PIGLIT_REPLAY_GFXRECON_INFO_BINARY',
            ('replay', 'gfxrecon-info_bin'),
            default='gfxrecon-info')
        cmd = [gfxrecon_info_bin, self._trace_path]
        ret = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=sys.stderr)
        lines = ret.stdout.decode(errors='replace').splitlines()
        print(ret.stdout.decode(errors='replace'))
        try:
            frames = re.search(_TOTAL_FRAMES_RE, lines[2])
            return int(frames.group(1))
        except Exception:
            return -1

    def _check_version(self, gfxrecon_replay_bin):
        cmd = [gfxrecon_replay_bin, '--version']
        ret = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=sys.stderr)
        lines = ret.stdout.decode(errors='replace').splitlines()
        print(ret.stdout.decode(errors='replace'))
        try:
            v = re.search(_VERSION_RE, lines[1])
            current = version.Version(v.group(1))
        except Exception:
            raise DumpBackendError(
                '[dump_trace_images] Unable to check the current '
                'gfxrecon-replay version.')
        if _MIN_VERSION > current:
            raise DumpBackendError(
                '[dump_trace_images] The current gfxrecon-replay version '
                'is {}. Try to update, at least to the {} version.'.format(
                    current, _MIN_VERSION))

    @dump_handler
    def dump(self):
        from PIL import Image
        outputprefix = path.join(self._output_dir,
                                 path.basename(self._trace_path))
        gfxrecon_replay_bin = core.get_option(
            'PIGLIT_REPLAY_GFXRECON_REPLAY_BINARY',
            ('replay', 'gfxrecon-replay_bin'),
            default='gfxrecon-replay')
        self._check_version(gfxrecon_replay_bin)
        if not self._calls:
            self._calls = [str(self._get_last_frame_call())]
        gfxrecon_replay_extra_args = core.get_option(
            'PIGLIT_REPLAY_GFXRECON_REPLAY_EXTRA_ARGS',
            ('replay', 'gfxrecon-replay_extra_args'),
            default='').split()
        cmd = ([gfxrecon_replay_bin] + gfxrecon_replay_extra_args +
               ['--screenshots', ','.join(self._calls),
                '--screenshot-dir', self._output_dir,
                self._trace_path])
        self._run_logged_command(cmd, None)
        for c in self._calls:
            bmp = '{}_frame_{}.bmp'.format(path.join(self._output_dir,
                                                     'screenshot'),
                                           c)
            outputfile = '{}-{}.png'.format(outputprefix, c)
            print('Writing: {} to {}'.format(bmp, outputfile))
            Image.open(bmp).save(outputfile)
            os.remove(bmp)


REGISTRY = Registry(
    extensions=['.gfxr'],
    backend=GFXReconstructBackend,
)



# coding=utf-8
#
# Copyright (c) 2015-2016, 2019 Intel Corporation
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import argparse
from typing import Any


class FileContentType(argparse.FileType):
    """A FileType type argument which reads the file content
    and closes the file"""

    def __init__(self, bufsize=-1, encoding=None, errors=None):
        super().__init__(
            mode="r", bufsize=bufsize, encoding=encoding, errors=errors
        )

    def __call__(self, string: str) -> Any:
        with super().__call__(string) as open_file:
            return open_file.read()


DEVICE = argparse.ArgumentParser(add_help=False)
DEVICE.add_argument(
    '-d', '--device-name',
    dest='device_name',
    required=False,
    default=None,
    help='the name of the graphics device used to replay traces')

KEEP_IMAGE = argparse.ArgumentParser(add_help=False)
KEEP_IMAGE.add_argument(
    '-k', '--keep-image',
    dest='keep_image',
    action='store_true',
    help=('forces keeping the image generated '
          'for the comparison.'))

YAML = argparse.ArgumentParser(add_help=False)
YAML.add_argument(
    '-y', '--yaml-file',
    dest='yaml_file',
    required=True,
    type=argparse.FileType('r'),
    help=('the name of the traces YAML description file listing traces '
          'and their checksums for each device'))

DOWNLOAD_URL = argparse.ArgumentParser(add_help=False)
DOWNLOAD_URL.add_argument(
    '-u', '--download-url',
    dest='download_url',
    required=False,
    default=None,
    help=('the URL from which to download the files'))

DOWNLOAD_CACHING_PROXY_URL = argparse.ArgumentParser(add_help=False)
DOWNLOAD_CACHING_PROXY_URL.add_argument(
    '-c', '--download-caching-proxy-url',
    dest='download_caching_proxy_url',
    required=False,
    default=None,
    help=('the URL for the caching proxy to be used as a prefix '
          'for the URL provided via --download-url'))

DOWNLOAD_FORCE = argparse.ArgumentParser(add_help=False)
DOWNLOAD_FORCE.add_argument(
    '-w', '--force-download',
    dest='force_download',
    action='store_true',
    help=('forces downloading '
          'even if the destination file already exists'))

DOWNLOAD_MINIO_HOST = argparse.ArgumentParser(add_help=False)
DOWNLOAD_MINIO_HOST.add_argument(
    '-m', '--minio_host',
    dest='download_minio_host',
    required=False,
    default=None,
    help=('name of MinIO server from which to download traces'))

DOWNLOAD_MINIO_BUCKET = argparse.ArgumentParser(add_help=False)
DOWNLOAD_MINIO_BUCKET.add_argument(
    '-b', '--minio_bucket',
    dest='download_minio_bucket',
    required=False,
    default=None,
    help=('Name of bucket in MinIO server containing the traces'))

DOWNLOAD_ROLE_SESSION_NAME = argparse.ArgumentParser(add_help=False)
DOWNLOAD_ROLE_SESSION_NAME.add_argument(
    '-r', '--role-session-name',
    dest='download_role_session_name',
    required=False,
    default=None,
    help=('role session name for authentication with MinIO'))

DOWNLOAD_JWT = argparse.ArgumentParser(add_help=False)
download_jwt_group = DOWNLOAD_JWT.add_mutually_exclusive_group()
download_jwt_group.add_argument(
    '-j', '--jwt',
    dest='download_jwt',
    required=False,
    default=None,
    help=('JWT token for authentication with MinIO'))
download_jwt_group.add_argument(
    "--jwt-file",
    dest="download_jwt",
    required=False,
    default=None,
    type=FileContentType(),
    help=("File containing JWT token for authentication with MinIO"),
)

DB_PATH = argparse.ArgumentParser(add_help=False)
DB_PATH.add_argument(
    '-p', '--db-path',
    dest='db_path',
    required=False,
    default='./replayer-db/',
    help=('the path to the objects db or where it will be created. '
          'Defaults to "./replayer-db/".'))

RESULTS_PATH = argparse.ArgumentParser(add_help=False)
RESULTS_PATH.add_argument(
    '-o', '--output',
    dest='output',
    required=False,
    default='./results/',
    help=('the path in which to place the results. Defaults to "./results/"'))



# coding=utf-8
#
# Copyright © 2021 Collabora Ltd.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Module provides a base class for Tests """

import glob
import os
import sys
import json

from framework import core, options
from framework import status
from .base import Test, WindowResizeMixin, ValgrindMixin, TestIsSkip


__all__ = [
    'PiglitCLTest',
    'PiglitGLTest',
    'PiglitBaseTest',
    'PiglitReplayerTest',
    'VkRunnerTest',
    'CL_CONCURRENT',
    'ROOT_DIR',
    'TEST_BIN_DIR',
]

if 'PIGLIT_BUILD_DIR' in os.environ:
    ROOT_DIR = os.environ['PIGLIT_BUILD_DIR']
else:
    ROOT_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '../..'))

TEST_BIN_DIR = os.path.normpath(os.path.join(ROOT_DIR, 'bin'))

CL_CONCURRENT = bool(not sys.platform.startswith('linux') or
                     glob.glob('/dev/dri/render*'))


class PiglitBaseTest(ValgrindMixin, Test):
    """
    PiglitTest: Run a "native" piglit test executable

    Expect one line prefixed PIGLIT: in the output, which contains a result
    dictionary. The plain output is appended to this dictionary
    """
    def __init__(self, command, run_concurrent=True, **kwargs):
        super(PiglitBaseTest, self).__init__(command, run_concurrent, **kwargs)

    @Test.command.getter
    def command(self):
        command = super(PiglitBaseTest, self).command

        def fixup_bin_path(c):
            # Prepend TEST_BIN_DIR to the path.
            if c == self._command[0]:
                return os.path.join(TEST_BIN_DIR, c)
            else:
                return c

        return [fixup_bin_path(c) for c in command]

    def interpret_result(self):
        out = []

        for each in self.result.out.split('\n'):
            if each.startswith('PIGLIT:'):
                deserial = json.loads(each[8:])
                if 'enumerate subtests' in deserial:
                    for n in deserial['enumerate subtests']:
                        self.result.subtests[n] = status.NOTRUN
                else:
                    self.result.update(deserial)
            else:
                out.append(each)

        self.result.out = '\n'.join(out)

        # XXX: There are a number of tests that now enumerate their subtests,
        # but don't properly report skip for all of them when the skip due to a
        # missing feature. We should fix these tests to do the right thing, but
        # for the moment this work around will suffice to keep things running.
        if self.result.raw_result is status.SKIP and self.result.subtests:
            for k, v in self.result.subtests.items():
                if v is status.NOTRUN:
                    self.result.subtests[k] = status.SKIP

        super(PiglitBaseTest, self).interpret_result()


class PiglitGLTest(WindowResizeMixin, PiglitBaseTest):
    """ OpenGL specific Piglit test class

    This Subclass provides provides an is_skip() implementation that skips glx
    tests on non-glx platforms

    This class also provides two additional keyword arguments, require_platform
    and exclude_platforms. require_platforms may be set to a list of platforms
    which the test requires to run. This should be resereved for platform
    specific tests, such as GLX specific tests, or EGL specific tests. Multiple
    platforms are allowed because EGL can be fulfilled by multiple platforms.
    exclude_platforms is a list of platforms a test should not be run on, this
    is useful for tests that are valid on more than one platform, but not on
    all of them. This will probably be mainly used to exclude gbm. These
    options are mutually exclusive.

    """
    def __init__(self, command, require_platforms=None, exclude_platforms=None,
                 **kwargs):
        # TODO: There is a design flaw in python2, keyword args can be
        # fulfilled as positional arguments. This sounds really great, until
        # you realize that because of it you cannot use the splat operator with
        # args and create new keyword arguments.
        # What we really want is __init__(self, *args, new_arg=None, **kwargs),
        # but this doesn't work in python2. In python3 thanks to PEP3102, you
        # can in fact do just that
        # The work around is to explicitly pass the arguments down.
        super(PiglitGLTest, self).__init__(command, **kwargs)

        assert not (require_platforms and exclude_platforms)

        if not require_platforms or set(require_platforms).issubset(
                set(core.PLATFORMS)):
            self.require_platforms = require_platforms or []
        else:
            raise Exception("Error: require_platform is not valid")

        if (not exclude_platforms or
                set(exclude_platforms).issubset(set(core.PLATFORMS))):
            self.exclude_platforms = exclude_platforms or []
        else:
            raise Exception("Error: exclude_platforms is not valid")

    def is_skip(self):
        """ Native Piglit-test specific skip checking

        If the platform for the run doesn't support glx (either directly as
        glx or through the hybrid glx/x11_egl setup that is default), then skip
        any glx specific tests.

        """
        platform = options.OPTIONS.env['PIGLIT_PLATFORM']
        if self.require_platforms and platform not in self.require_platforms:
            raise TestIsSkip(
                'Test requires one of the following platforms "{}" '
                'but the platform is "{}"'.format(
                    self.require_platforms, platform))
        elif self.exclude_platforms and platform in self.exclude_platforms:
            raise TestIsSkip(
                'Test cannot be run on any of the following platforms "{}" '
                'and the platform is "{}"'.format(
                    self.exclude_platforms, platform))
        super(PiglitGLTest, self).is_skip()

    @PiglitBaseTest.command.getter
    def command(self):
        """ Automatically add -auto and -fbo as appropriate """
        if not self.run_concurrent:
            return self.keys() + super(PiglitGLTest, self).command + ['-auto']
        else:
            return self.keys() + super(PiglitGLTest, self).command + ['-auto', '-fbo']

    @command.setter
    def command(self, new):
        self._command = [n for n in new if n not in ['-auto', '-fbo']]


class ASMParserTest(PiglitBaseTest):

    """Test class for ASM parser tests."""

    def __init__(self, type_, filename, env=None):
        super(ASMParserTest, self).__init__(['asmparsertest', type_], env=env)
        self.filename = filename

    @PiglitBaseTest.command.getter
    def command(self):
        command = super(ASMParserTest, self).command
        return self.keys() + command + [os.path.join(ROOT_DIR, self.filename)]


class BuiltInConstantsTest(PiglitBaseTest):

    """Test class for handling built in constants tests."""

    @PiglitBaseTest.command.getter
    def command(self):
        command = super(BuiltInConstantsTest, self).command

        command[1] = os.path.join(ROOT_DIR, 'tests', command[1])
        return self.keys() + command; 


class PiglitCLTest(PiglitBaseTest):  # pylint: disable=too-few-public-methods
    """ OpenCL specific Test class.

    Set concurrency based on CL requirements.

    """
    def __init__(self, command, run_concurrent=CL_CONCURRENT, **kwargs):
        if self.timeout is None:
            self.timeout = 60
        super(PiglitCLTest, self).__init__(command, run_concurrent, **kwargs)


class CLProgramTester(PiglitCLTest):

    """Class for cl-program-tester tests."""

    def __init__(self, filename, **kwargs):
        super(CLProgramTester, self).__init__(['cl-program-tester'], **kwargs)
        self.filename = filename

    @PiglitCLTest.command.getter
    def command(self):
        command = super(CLProgramTester, self).command
        return self.keys() + command + [os.path.join(ROOT_DIR, self.filename)]


class VkRunnerTest(PiglitBaseTest):
    """ Make a PiglitTest instance for a VkRunner shader test file """

    def __init__(self, filename, env=None):
        vkrunner_bin = core.get_option('PIGLIT_VKRUNNER_BINARY',
                                       ('vkrunner', 'bin'),
                                       default='vkrunner')

        super(VkRunnerTest, self).__init__(
            [vkrunner_bin],
            run_concurrent=True,
            env=env)

        self.filename = filename

    @PiglitBaseTest.command.getter
    def command(self):
        # self._command is used because we don't want PiglitBaseTest
        # to prepend TEST_BIN_DIR so that it will look for vkrunner in
        # the search path.
        return self.keys() + self._command + [os.path.join(ROOT_DIR, self.filename)]


class PiglitReplayerTest(PiglitBaseTest):
    """ Make a PiglitTest instance for a Replayer test

    Runs a single replayer test.

    Arguments:
    extra_args -- to pass to replayer

    """

    RESULTS_PATH = None

    def __init__(self, subcommand, extra_args, **kwargs):
        super(PiglitReplayerTest, self).__init__(
            ['replayer.py', subcommand, 'trace'], **kwargs)
        self.extra_args = extra_args

    @PiglitBaseTest.command.getter
    def command(self):
        command = super(PiglitReplayerTest, self).command
        if self.RESULTS_PATH is not None:
            command += ['--output', self.RESULTS_PATH]
        return self.keys() + command + self.extra_args

    def interpret_result(self):
        # Python's unhandled exceptions use "1" as exit code value.  We want to
        # interpret this as a CRASH. Therefore, we change the exit code value
        # to a negative one, which is the one the Test base class interprets as
        # a CRASH.
        if self.result.returncode == 1:
            self.result.returncode = -1

        super(PiglitReplayerTest, self).interpret_result()



#!/usr/local/bin/python3.13
# coding=utf-8
#
# Copyright (c) 2019 Collabora Ltd
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import argparse
import atexit
import os
import shutil
import tempfile
from pathlib import Path


def cleanup(dirpath):
    shutil.rmtree(dirpath)


TMP_DIR = tempfile.mkdtemp()
atexit.register(cleanup, TMP_DIR)
RENDERDOC_DEBUG_FILE = TMP_DIR + "/renderdoc.log"

# Needs to be in the environment before importing the module
os.environ['RENDERDOC_DEBUG_LOG_FILE'] = RENDERDOC_DEBUG_FILE
import renderdoc as rd  # noqa: E402

try:
    from renderdoc import ResultCode
except ImportError:
    from renderdoc import ReplayStatus as ResultCode


def find_draw_with_event_id(controller, event_id):
    try:
        for root_action in controller.GetRootActions():
            if root_action.eventId == event_id:
                return root_action
    except AttributeError:
        for root_drawcall in controller.GetDrawcalls():
            if root_drawcall.eventId == event_id:
                return root_drawcall

    return None


def dump_image(controller, event_id, output_dir, tracefile):
    draw = find_draw_with_event_id(controller, event_id)
    if draw is None:
        raise RuntimeError("Couldn't find draw call with event ID " +
                           str(event_id))

    controller.SetFrameEvent(draw.eventId, True)

    texsave = rd.TextureSave()

    # Select the first color output
    texsave.resourceId = draw.outputs[0]

    if texsave.resourceId == rd.ResourceId.Null():
        return

    filepath = Path(output_dir)
    filepath.mkdir(parents=True, exist_ok=True)
    filepath = filepath / (tracefile + "-" + str(int(draw.eventId)) + ".png")

    try:
        filename: str = draw.GetName(controller.GetStructuredFile())
    except AttributeError:
        filename: str = draw.name

    print(f"Saving image at event ID {draw.eventId}: {filename} to {filepath}")

    # Most formats can only display a single image per file, so we select the
    # first mip and first slice
    texsave.mip = 0
    texsave.slice.sliceIndex = 0

    # For formats with an alpha channel, preserve it
    texsave.alpha = rd.AlphaMapping.Preserve
    texsave.destType = rd.FileType.PNG
    controller.SaveTexture(texsave, str(filepath))


def load_capture(filename):
    cap = rd.OpenCaptureFile()

    status = cap.OpenFile(filename, '', None)

    if status != ResultCode.Succeeded:
        raise RuntimeError("Couldn't open file: " + str(status))
    if not cap.LocalReplaySupport():
        raise RuntimeError("Capture cannot be replayed")

    status, controller = cap.OpenCapture(rd.ReplayOptions(), None)

    if status != ResultCode.Succeeded:
        if os.path.exists(RENDERDOC_DEBUG_FILE):
            print(open(RENDERDOC_DEBUG_FILE, "r").read())
        raise RuntimeError("Couldn't initialise replay: " + str(status))

    if os.path.exists(RENDERDOC_DEBUG_FILE):
        open(RENDERDOC_DEBUG_FILE, "w").write("")

    return (cap, controller)


def renderdoc_dump_images(filename, event_ids, output_dir):
    rd.InitialiseReplay(rd.GlobalEnvironment(), [])
    cap, controller = load_capture(filename)

    tracefile = Path(filename).name

    if not event_ids:
        try:
            event_ids.append(controller.GetRootActions()[-1].eventId)
        except AttributeError:
            event_ids.append(controller.GetDrawcalls()[-1].eventId)

    for event_id in event_ids:
        dump_image(controller, event_id, output_dir, tracefile)

    cap.Shutdown()

    rd.ShutdownReplay()


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('file_path',
                        help='the path to a trace file.')
    parser.add_argument('output_dir',
                        help='the path in which to place the results')
    parser.add_argument('draw_id',
                        type=int,
                        nargs=argparse.REMAINDER,
                        help=('a draw-id number from the trace to dump. '
                              'If none are provided, by default, '
                              'the last frame will be used.'))

    args = parser.parse_args()

    renderdoc_dump_images(args.file_path, args.draw_id, args.output_dir)


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.



# coding=utf-8
# Copyright (c) 2014-2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module providing utility functions to work with piglit groups.

Instead of using posixpath (or the generic os.path) for working with tests this
module should be preferred.

Piglit groups look much like posix paths, they are '/' delimited with each
element representing a group, and the final element being the test name. Unlike
posix paths they may not start with a leading '/'.

"""

__all__ = [
    'SEPARATOR',
    'commonprefix',
    'format',
    'from_path',
    'groupname',
    'join',
    'split',
    'splitname',
    'testname',
]

SEPARATOR = '@'


def testname(group):
    """Return the last element of a group name.

    Provided the value 'group1|group2|test1' will provide 'test1', this
    does not enforce any rules that the final element is a test name, and can
    be used to shaved down groups.

    Analogous to os.path.basename

    """
    return splitname(group)[1]


def groupname(group):
    """Return all groups except the last.

    Provided the value 'group1/group2/test1' will provide 'group1/group2', this
    does not enforce any rules that the final element is a test name, and can
    be used to shaved down groups.

    Analogous to os.path.dirname

    """
    return splitname(group)[0]


def splitname(group):
    """Split a group name, Returns tuple "(group, test)"."""
    i = group.rfind(SEPARATOR) + 1
    head, tail = group[:i], group[i:]
    head = head.rstrip(SEPARATOR)

    return head, tail


def commonprefix(args):
    """Given a list of groups, returns the longest common leading component."""
    if len(args) == 1:
        return args[0]
    elif any(e == '' for e in args):
        return ''

    common = []

    for elems in zip(*[split(a) for a in args]):
        iter_ = iter(elems)
        first = next(iter_)
        if all(first == r for r in iter_):
            common.append(first)
        else:
            break

    # Join needs at least one element to join
    if common:
        return join(*common)
    else:
        return ''


def join(first, *args):
    """Join multiple groups together.

    This function is implemented via string concatenation, while most
    pythonistas would use list joining, because it is accepted as better.  I
    wrote a number of implementations and timed them with timeit.  I found for
    small number of joins (2-10) that str concatenation was quite a bit faster,
    at around 100 elements list joining became faster. Since most of piglit's
    use of join is for 2-10 elements I used string concatenation, which is
    coincidentally very similar to the way posixpath.join is implemented.

    """
    # If first happens to be a non-existent value, walk through args until we
    # find a real value and use that.
    args = (a for a in args)
    if not first:
        for group in args:
            if group:
                first = group
                break

    for group in args:
        # Only append things if the group is not empty, otherwise we'll get
        # extra SEPARATORs where we don't want them
        if group:
            if not first.endswith(SEPARATOR):
                first += SEPARATOR
            first += group

    return first


def split(group):
    """Split the group into a list of elements.

    If input is '' return an empty list

    """
    if group == '':
        return []
    return group.split(SEPARATOR)


def from_path(path):
    """Create a group from a path.

    This function takes a path, and creates a group out of it.

    This safely handles both Windows and Unix style paths.

    """
    assert isinstance(path, str), 'Type must be str'

    if '\\' in path:
        path = path.replace('\\', SEPARATOR)
    if '/' in path:
        path = path.replace('/', SEPARATOR)
    if '.' == path:
        return ''
    return path


def format(name):
    """Format an internal name for printing.

    It doesn't matter how the name is stored internally, presenting a
    consistence, clean interface on the command line that doesn't contain any
    ugly or problematic characters is important.

    This replaces SEPARATOR with '/', which is what most devs are used to and
    want to see.

    """
    assert isinstance(name, str)
    return name.replace(SEPARATOR, '/')



#!/usr/local/bin/python3.13
# coding=utf-8

one_parameter_110 = [
    "radians",
    "degrees",
    "sin",
    "cos",
    "tan",
    "asin",
    "acos",
    "atan",
    "exp",
    "log",
    "exp2",
    "log2",
    "sqrt",
    "inversesqrt",
    "abs",
    "sign",
    "floor",
    "ceil",
    "fract",
    "length",
    "normalize",
    ]

one_parameter_110_fs = [
    "dFdx",
    "dFdy",
    "fwidth",
    ]

two_parameter_110 = [
    "atan",
    "pow",
    "mod",
    "min",
    "max",
    "step",
    "distance",
    "dot",
    "reflect",
    ]

three_parameter_110 = [
    "clamp",
    "mix",
    "smoothstep",
    "faceforward",
    ]

misc_parameter_110 = [
    ["genType", "mod", "genType", "float"],
    ["genType", "min", "genType", "float"],
    ["genType", "max", "genType", "float"],
    ["genType", "clamp", "genType", "float", "float"],
    ["genType", "mix", "genType", "genType", "float"],
    ["genType", "step", "float", "genType"],
    ["genType", "smoothstep", "float", "float", "genType"],
    ["genType", "refract", "genType", "genType", "float"],
    ["float", "noise1", "genType"],
    ["vec2", "noise2", "genType"],
    ["vec3", "noise3", "genType"],
    ["vec4", "noise4", "genType"],
    ]

vec_parameter_110 = [
    ["bvec", "lessThan", "vecType", "vecType"],
    ["bvec", "lessThanEqual", "vecType", "vecType"],
    ["bvec", "greaterThan", "vecType", "vecType"],
    ["bvec", "greaterThanEqual", "vecType", "vecType"],
    ["bvec", "equal", "vecType", "vecType"],
    ["bvec", "notEqual", "vecType", "vecType"],
    ]

bvec_parameter_110 = [
    ["bvec", "equal", "bvec", "bvec"],
    ["bvec", "notEqual", "bvec", "bvec"],
    ["bool", "any", "bvec"],
    ["bool", "all", "bvec"],
    ["bvec", "not", "bvec"],
    ]

genType = ["float", "vec2", "vec3", "vec4"]
vecType = ["vec2", "vec3", "vec4", "ivec2", "ivec3", "ivec4"]
bvecType = ["bvec2", "bvec3", "bvec4"]


def typename(t, gen):
    if (t == "genType"):
        return gen
    elif (t == "vecType"):
        return gen
    elif (t == "bvec"):
        return "bvec" + gen[len(gen) - 1]
    else:
        return t


def emit_pattern_call(f, pat_list, type_list):
    for sig in pat_list:
        for t in type_list:
            name = sig[1]
            ret = typename(sig[0], t)

            params = "u_%s" % (typename(sig[2], t))
            for p in sig[3:]:
                params += ", u_%s" % (typename(p, t))

            f.write("  t_%s = %s(%s);\n" % (ret, name, params))

    f.write("\n")


def emit_matrixCompMult(f, matTypes):
    for mat in matTypes:
        f.write("  t_%s = matrixCompMult(u_%s, u_%s);\n" % (mat, mat, mat))


def emit_110_tests(f):
    f.write("/* PASS */\n#version 110\n\n")
    f.write("uniform float u_float;\n")

    for s in ["vec", "ivec", "bvec", "mat"]:
        for i in [2, 3, 4]:
            f.write("uniform %s%d u_%s%d;\n" % (s, i, s, i))

    f.write("""\nvoid main()
{
  gl_Position = gl_Vertex;

  float t_float = float(0.0);
  bool t_bool = false;\n""")

    for i in [2, 3, 4]:
        f.write("  vec%d  t_vec%d  = vec%d (0.0);\n" % (i, i, i))
        f.write("  bvec%d t_bvec%d = bvec%d(0.0);\n" % (i, i, i))
        f.write("  mat%d  t_mat%d  = mat%d (0.0);\n" % (i, i, i))

    f.write("\n")

    for name in one_parameter_110:
        for t in genType:
            f.write("  t_%s += %s(u_%s);\n" % (t, name, t))

    f.write("\n")

    for name in two_parameter_110:
        for t in genType:
            f.write("  t_%s += %s(u_%s, -u_%s);\n" % (t, name, t, t))

    f.write("\n")

    for name in three_parameter_110:
        for t in genType:
            f.write("  t_%s += %s(u_%s, -u_%s, 11.7 * u_%s);\n" %
                    (t, name, t, t, t))

    f.write("\n")

    emit_pattern_call(f, misc_parameter_110, genType)
    emit_pattern_call(f, vec_parameter_110, vecType)
    emit_pattern_call(f, bvec_parameter_110, bvecType)
    emit_matrixCompMult(f, ["mat2", "mat3", "mat4"])

    f.write("}\n")


def emit_110_fs_tests(f):
    f.write("/* PASS */\n#version 110\n\n")
    f.write("uniform float u_float;\n")

    for i in [2, 3, 4]:
        f.write("uniform vec%d u_vec%d;\n" % (i, i))

    f.write("""\nvoid main()
{
  gl_FragColor = gl_Color;

  float t_float = float(0.0);\n""")

    for i in [2, 3, 4]:
        f.write("  vec%d  t_vec%d  = vec%d (0.0);\n" % (i, i, i))

    f.write("\n")

    for name in one_parameter_110_fs:
        for t in genType:
            f.write("  t_%s = %s(u_%s);\n" % (t, name, t))

    f.write("}\n")


def emit_120_tests(f):
    f.write("/* PASS */\n#version 120\n\n")

    for cols in [2, 3, 4]:
        f.write("uniform mat%d u_mat%d;\n" % (cols, cols))
        f.write("uniform vec%d u_vec%d;\n" % (cols, cols))

        for rows in [2, 3, 4]:
            mat = "mat%dx%d" % (cols, rows)
            f.write("uniform %s u_%s;\n" % (mat, mat))

    f.write("""\nvoid main()
{
  gl_Position = gl_Vertex;

""")

    for cols in [2, 3, 4]:
        f.write("  mat%d   t_mat%d   = mat%d  (0.0);\n" % (cols, cols, cols))

        for rows in [2, 3, 4]:
            mat = "mat%dx%d" % (cols, rows)
            f.write("  %s t_%s = %s(0.0);\n" % (mat, mat, mat))

    f.write("\n")

    for cols in [2, 3, 4]:
        dst = "t_mat%d" % (cols)
        src = "u_mat%d" % (cols)
        f.write("  %s   = transpose(%s);\n" % (dst, src))

        for rows in [2, 3, 4]:
            dst = "t_mat%dx%d" % (cols, rows)
            src = "u_mat%dx%d" % (rows, cols)
            f.write("  %s = transpose(%s);\n" % (dst, src))

    f.write("\n")

    for cols in [2, 3, 4]:
        f.write("  t_mat%d   = outerProduct(u_vec%d, u_vec%d);\n"
                % (cols, cols, cols))

        for rows in [2, 3, 4]:
            f.write("  t_mat%dx%d = outerProduct(u_vec%d, u_vec%d);\n" %
                    (cols, rows, rows, cols))

    f.write("\n")

    emit_matrixCompMult(f, ["mat2x2", "mat2x3", "mat2x4",
                            "mat3x2", "mat3x3", "mat3x4",
                            "mat4x2", "mat4x3", "mat4x4"])

    f.write("}\n")


f = open("builtin-functions-110.vert", "w")
emit_110_tests(f)
f.close()

f = open("builtin-functions-110.frag", "w")
emit_110_fs_tests(f)
f.close()

f = open("builtin-functions-120.vert", "w")
emit_120_tests(f)
f.close()



# coding=utf-8
# Copyright (c) 2015-2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Stores global piglit options.

This is as close to a true global function as python gets. The only deal here
is that while you can mutate

"""

import os

__all__ = ['OPTIONS']

# pylint: disable=too-few-public-methods


class _Options(object):  # pylint: disable=too-many-instance-attributes
    """Contains all options for a piglit run.

    This is used as a sort of global state object, kinda like piglit.conf. This
    big difference here is that this object is largely generated internally,
    and is controlled mostly through command line options rather than through
    the configuration file.

    Options are as follows:
    execute -- False for dry run
    valgrind -- True if valgrind is to be used
    env -- environment variables set for each test before run
    deqp_mustpass -- True to enable the use of the deqp mustpass list feature.
    """

    def __init__(self):
        self.execute = True
        self.valgrind = False
        self.sync = False
        self.deqp_mustpass = False
        self.process_isolation = True
        self.jobs = None
        self.force_glsl = False

        # env is used to set some base environment variables that are not going
        # to change across runs, without sending them to os.environ which is
        # fickle and easy to break
        self.env = {
            'PIGLIT_SOURCE_DIR':
                os.environ.get(
                    'PIGLIT_SOURCE_DIR',
                    os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                 '..')))
        }

    def clear(self):
        """Reinitialize all values to defaults."""
        self.__init__()

    def __iter__(self):
        for key, values in self.__dict__.items():
            if not key.startswith('_'):
                yield key, values


OPTIONS = _Options()



# coding=utf-8
# Copyright (c) 2016, 2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""Module implementing classes for monitoring

This module provides a set of classes for monitoring the system on multiple
sources. The sources can be standard files, locked files or the dmesg.
The patterns of the errors are defined in piglit.conf, at the section
monitored-errors.

When one of the regex is found in the corresponding source Piglit will abort
with code 3.

"""

import abc
import errno
import os
import re

from framework.core import PIGLIT_CONFIG
from framework.dmesg import LinuxDmesg
from framework import exceptions

__all__ = [
    'BaseMonitoring',
    'Monitoring',
    'MonitoringFile',
    'MonitoringLinuxDmesg',
]



class Monitoring(object):
    """Class that initialize and update monitoring classes

    This reads the rules from the section monitored-errors in piglit.conf
    and initializes the monitoring objects. Their type must be derived from
    BaseMonitoring and specialized according the field 'type'.

    """

    # starting time: user must know current machine/GPU state before
    # starting piglit: we can resume test execution based on last stopping
    # if we consider that our system is still in good shape (ie not blocking
    # run forever)
    _abort_error = None
    _monitoring_rules = None

    def __init__(self, monitoring_enabled):
        """Create a LinuxMonitored instance"""
        # Get the monitoring rules from piglit.conf and store them into a dict.
        self._monitoring_rules = {}

        if monitoring_enabled and PIGLIT_CONFIG.has_section('monitored-errors'):
            for key, _ in PIGLIT_CONFIG.items('monitored-errors'):
                if PIGLIT_CONFIG.has_section(key):
                    type = PIGLIT_CONFIG.required_get(key, 'type')
                    regex = PIGLIT_CONFIG.required_get(key, 'regex')
                    parameters = PIGLIT_CONFIG.required_get(key, 'parameters')

                    self.add_rule(key, type, parameters, regex)

    @property
    def abort_needed(self):
        """Simply return if _abort_error variable is not empty"""
        return self._abort_error is not None

    @property
    def error_message(self):
        """Simply return _abort_error message"""
        return self._abort_error

    def add_rule(self, key, type, parameters, regex):
        """Add a new monitoring rule

        This method adds a new monitoring rule. The type must be file,
        locked_file or dmesg.

        Arguments:
        key -- The key for storing the rule in a dict
        type -- The type of monitoring
        parameters -- Depending on the type, can be a file path or a command
                      options
        regex -- The rule regex

        """
        rule = None

        if type == 'file':
            rule = MonitoringFile(parameters,
                                  regex)
        elif type == 'locked_file':
            rule = MonitoringFile(parameters,
                                  regex, True)
        elif type == 'dmesg':
            rule = MonitoringLinuxDmesg(parameters,
                                        regex)
        else:
            raise exceptions.PiglitFatalError(
                "No available monitoring class for the type {}.".format(type))

        self._monitoring_rules[key] = rule

    def delete_rule(self, key):
        """Remove a monitoring rule

        Arguments:
        key -- The rule key

        """
        self._monitoring_rules.pop(key, None)

    def update_monitoring(self):
        """Update the new messages for each monitoring object"""
        if self._monitoring_rules:
            for monitoring_rule in self._monitoring_rules.values():
                monitoring_rule.update_monitoring()

    def check_monitoring(self):
        """Check monitoring objects statue

        This method checks the state for each monitoring object.
        If one of them found the pattern in the new messages,
        set itself on abort_needed state.

        """
        # Get a new snapshot of the source
        self.update_monitoring()

        if self._monitoring_rules:
            for rule_key, monitoring_rule in self._monitoring_rules.items():
                # Get error message
                self._abort_error = monitoring_rule.check_monitoring()
                # if error message is not empty, abort is requested
                if self.abort_needed:
                    self._abort_error = "From the rule {}:\n{}".format(
                        rule_key,
                        self._abort_error)
                    break


class BaseMonitoring(metaclass=abc.ABCMeta):
    """Abstract base class for Monitoring derived objects

    This provides the bases of the constructor and most subclasses should call
    super() in __init__(). It provides a concrete implementation of the
    check_monitoring() method, which should be suitible for all subclasses.

    The update_monitoring() method need to be override for all subclasses.

    """
    @abc.abstractmethod
    def __init__(self, monitoring_source, regex):
        """Abstract constructor for BaseMonitoring subclasses

        Arguments;
        monitoring_source -- The source to monitor
        regex -- The errors pattern

        """
        self._monitoring_source = monitoring_source
        self._monitoring_regex = re.compile(regex)
        self._new_messages = ''

    @property
    def new_messages(self):
        """Simply return if _abort_error variable is not empty"""
        return self._new_messages

    @abc.abstractmethod
    def update_monitoring(self):
        """Update _new_messages list

        This method should read a monitoring source like a file or a program
        output, determine new entries since the last read of the source,
        and update self._new_messages with those entries

        """
        pass

    def check_monitoring(self):
        """Check _new_messages

        This method checks if the regex is found in the new messages,
        then return a string that contain the monitored location and
        the matched line

        """
        if self._new_messages and self._monitoring_regex:
            for line in self._new_messages:
                if self._monitoring_regex.search(line):
                    return line


if os.name == 'posix':
    import fcntl


    class MonitoringFile(BaseMonitoring):
        """Monitoring from a file

        This class is for monitoring the system from a file that
        can be a standard file or a locked file. The locked file
        algorithm uses fnctl, so it doesn't work on non Unix systems

        Arguments:
        is_locked -- True if the target is a locked file

        """
        _is_locked = False

        def __init__(self, monitoring_source, regex, is_locked=False):
            """Create a MonitoringFile instance"""
            self._last_message = None
            self._is_locked = is_locked
            super(MonitoringFile, self).__init__(monitoring_source, regex)

        def update_monitoring(self):
            """Open the file and get the differences

            Get the contents of the file, then calculate new messages.
            This implements also a specific method for reading locked files.

            """

            try:
                with open(self._monitoring_source, 'r') as f:
                    lines = []
                    if self._is_locked:
                        # Create a duplicated file descriptor, this avoid lock
                        fd = os.dup(f.fileno())
                        # use I/O control for reading the lines
                        fcntl.fcntl(fd, fcntl.F_SETFL, os.O_NONBLOCK)

                        while True:
                            try:
                                line = os.read(fd, 1024)
                                if not line:
                                    break;
                            except OSError as e:
                                if e.errno == errno.EAGAIN:
                                    break
                                else:
                                    raise e
                            lines.append(line.decode("utf-8", "strict"))
                        os.close(fd)

                    else:
                        lines = f.read().splitlines()

                    f.close()

                    # Find all new entries, do this by slicing the list of the
                    # lines to only returns elements after the last element
                    # stored. If there are not matches a value error is raised,
                    # that means all of the lines are new
                    l = 0
                    for index, item in enumerate(reversed(lines)):
                        if item == self._last_message:
                            l = len(lines) - index  # don't include the matched element
                            break
                    self._new_messages = lines[l:]
                    # Attempt to store the last element of lines,
                    # unless there was no line
                    self._last_message = lines[-1] if lines else None
            except Exception:
                # if an error occurred, we consider there are no new messages
                self._new_messages = []
                pass


    class MonitoringLinuxDmesg(BaseMonitoring, LinuxDmesg):
        """Monitoring on dmesg

        This class is for monitoring on the system dmesg. It's inherited
        from LinuxDmesg for the dmesg processing methods.

        Work only on Linux operating system.

        """
        def __init__(self, monitoring_source, regex):
            """Create a MonitoringLinuxDmesg instance"""
            self.DMESG_COMMAND = ['dmesg']+monitoring_source.split()
            BaseMonitoring.__init__(self, monitoring_source, regex)
            LinuxDmesg.__init__(self)

        def update_monitoring(self):
            """Call update_dmesg"""
            self.update_dmesg()



# coding=utf-8
# Copyright (c) 2014-2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A collection of various bits that don't seem to belong anywhere else.

This is the classic catchall "utils" module from most projects, that for
historically reasons is called "core" in piglit.

"""

import configparser
import errno
import functools
import os
import subprocess
import time

from framework import exceptions

__all__ = [
    'PIGLIT_CONFIG',
    'PLATFORMS',
    'PiglitConfig',
    'collect_system_info',
    'get_option',
    'parse_listfile',
]

PLATFORMS = ["glx", "x11_egl", "wayland", "gbm", "mixed_glx_egl", "wgl", "surfaceless_egl"]


class PiglitConfig(configparser.ConfigParser):
    """Custom Config parser that provides a few extra helpers."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = None

    def read_file(self, f, source=None):
        super().read_file(f, source)
        self.filename = os.path.abspath(source or f.name)

    def safe_get(self, section, option, fallback=None, **kwargs):
        """A version of self.get that doesn't raise NoSectionError or
        NoOptionError.

        This is equivalent to passing if the option isn't found. It will return
        None if an error is caught

        """
        try:
            return self.get(section, option, **kwargs)
        except (configparser.NoOptionError, configparser.NoSectionError):
            return fallback

    def required_get(self, section, option, **kwargs):
        """A version of self.get that raises PiglitFatalError.

        If self.get returns NoSectionError or NoOptionError then this will
        raise a PiglitFatalException, aborting the program.

        """
        try:
            return self.get(section, option, **kwargs)
        except configparser.NoSectionError:
            raise exceptions.PiglitFatalError(
                'No Section "{}" in file "{}".\n'
                'This section is required.'.format(
                    section, self.filename))
        except configparser.NoOptionError:
            raise exceptions.PiglitFatalError(
                'No option "{}"  from section "{}" in file "{}".\n'
                'This option is required.'.format(
                    option, section, self.filename))


PIGLIT_CONFIG = PiglitConfig(allow_no_value=True)


def get_config(arg=None):
    if arg:
        PIGLIT_CONFIG.read_file(arg)
    else:
        # Load the piglit.conf. First try looking in the current directory,
        # then trying the XDG_CONFIG_HOME, then $HOME/.config/, finally try the
        # piglit root dir
        for d in ['.',
                  os.environ.get('XDG_CONFIG_HOME',
                                 os.path.expandvars('$HOME/.config')),
                  os.path.join(os.path.dirname(__file__), '..')]:
            try:
                with open(os.path.join(d, 'piglit.conf'), 'r') as f:
                    PIGLIT_CONFIG.read_file(f)
                break
            except IOError:
                pass


def get_option(env_varname, config_option, default=None, required=False):
    """Query the given environment variable and then piglit.conf for the option.

    Return the value of the default argument if opt is None.

    """
    opt = os.environ.get(env_varname, None)
    if opt is not None:
        return opt

    opt = PIGLIT_CONFIG.safe_get(config_option[0], config_option[1])

    if required and not opt:
        raise exceptions.PiglitFatalError(
            'Cannot get env "{}" or conf value "{}:{}"'.format(
                env_varname, config_option[0], config_option[1]))
    return opt or default


def check_dir(dirname, failifexists=False, handler=None):
    """Check for the existence of a directory and create it if possible.

    This function will check for the existence of a directory. If that
    directory doesn't exist it will try to create it. If the directory does
    exist than it does one of two things.
    1) If "failifexists" is False (default): it will just return
    2) If "failifexists" is True it will raise an PiglitException, it is the
    job of the caller using failifexists=True to handle this exception

    Both failifexists and handler can be passed, but failifexists will have
    precedence.

    Arguments:
    dirname -- the name of the directory to check

    Keyword Arguments:
    failifexists -- If True and the directory exists then PiglitException will
                    be raised (default: False)
    handler -- a callable that is passed dirname if the thing to check exists.

    """
    try:
        os.stat(dirname)
    except OSError as e:
        # If the error is not "no file or directory" or "not a dir", then
        # either raise an exception, call the handler function, or return
        if e.errno not in [errno.ENOENT, errno.ENOTDIR]:
            if failifexists:
                raise exceptions.PiglitException
            elif handler is not None:
                handler(dirname)

    try:
        # makedirs is expensive, so check before # calling it.
        if not os.path.exists(dirname):
            os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def collect_system_info():
    """ Get relevant information about the system running piglit

    This method runs through a list of tuples, where element 1 is the name of
    the program being run, and element 2 is a command to run (in a form accepted
    by subprocess.Popen)

    """
    progs = [('wglinfo', ['wglinfo']),
             ('glxinfo', ['glxinfo']),
             ('uname', ['uname', '-a']),
             ('clinfo', ['clinfo']),
             ('lspci', ['lspci', '-nn'])]

    result = {}

    for name, command in progs:
        try:
            out = subprocess.check_output(command, stderr=subprocess.STDOUT)
            result[name] = out.decode('utf-8')
        except OSError as e:
            # If we get the 'no file or directory' error then pass, that means
            # that the binary isn't installed or isn't relevant to the system.
            # If it's any other OSError, then raise
            if e.errno != errno.ENOENT and e.errno != errno.EACCES:
                raise
        except subprocess.CalledProcessError:
            # If the binary is installed by doesn't work on the window system
            # (glxinfo) it will raise this error. go on
            pass

    return result


def parse_listfile(filename):
    """
    Parses a newline-seperated list in a text file and returns a python list
    object. It will expand tildes on Unix-like system to the users home
    directory.

    ex file.txt:
        ~/tests1
        ~/tests2/main
        /tmp/test3

    returns:
        ['/home/user/tests1', '/home/users/tests2/main', '/tmp/test3']
    """
    with open(filename, 'r') as file:
        return [os.path.expanduser(i.strip()) for i in file.readlines()]


class lazy_property(object):  # pylint: disable=invalid-name,too-few-public-methods
    """Descriptor that replaces the function it wraps with the value generated.

    This makes a property that is truly lazy, it is calculated once on demand,
    and stored. This makes this very useful for values that you might want to
    calculate and reuse, but they cannot change.

    This works by very cleverly shadowing itself with the calculated value. It
    adds the value to the instance, which pushes itself up the MRO and will
    never be queried again.

    """
    def __init__(self, func):
        self.__func = func

    def __get__(self, instance, cls):
        value = self.__func(instance)
        setattr(instance, self.__func.__name__, value)
        return value


def timer_ms(func):
    """A decorator function that measures the runtime of a given function in
    milliseconds and prints the result.
    The function name and runtime in ms will be printed in the format
    "Finished [function name] in [runtime] ms".

    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time: float = time.monotonic()
        value = func(*args, **kwargs)
        end_time: float = time.monotonic()
        run_time: float = end_time - start_time
        print(f"Finished {func.__name__} in {run_time.__round__(6) * 1000} ms")
        return value

    return wrapper



# coding=utf-8
# Copyright (c) 2014-2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module like interface for backends.

This package provides an abstract interface for working with backends, which
implement various functions as provided in the Registry class, and then provide
a Registry instance as REGISTRY, which maps individual objects to objects that
piglit expects to use. This module then provides a thin abstraction layer so
that piglit is never aware of what backend it's using, it just asks for an
object and receives one.

Most consumers will want to import framework.backends and work directly with
the helper functions here. For some more detailed uses (test cases especially)
the modules themselves may be used.

When this module is loaded it will search through framework/backends for python
modules (those ending in .py), and attempt to import them. Then it will look
for an attribute REGISTRY in those modules, and if it as a
framework.register.Registry instance, it will add the name of that module (sans
.py) as a key, and the instance as a value to the BACKENDS dictionary. Each of
the helper functions in this module uses that dictionary to find the function
that a user actually wants.

"""

import os
import importlib

from .register import Registry
from .compression import COMPRESSION_SUFFIXES

__all__ = [
    'BACKENDS',
    'BackendError',
    'BackendNotImplementedError',
    'get_backend',
    'load',
    'set_meta',
    'write',
]


class BackendError(Exception):
    pass


class BackendNotImplementedError(NotImplementedError):
    pass


def _register():
    """Register backends.

    Walk through the list of backends and register them to a name in a
    dictionary, so that they can be referenced from helper functions.

    """
    registry = {}

    for module in os.listdir(os.path.dirname(os.path.abspath(__file__))):
        module, extension = os.path.splitext(module)
        if extension == '.py':
            mod = importlib.import_module('framework.backends.{}'.format(module))
            if hasattr(mod, 'REGISTRY') and isinstance(mod.REGISTRY, Registry):
                registry[module] = mod.REGISTRY

    return registry


BACKENDS = _register()


def get_backend(backend):
    """Returns a BackendInstance based on the string passed.

    If the backend isn't a known module, then a BackendError will be raised, it
    is the responsibility of the caller to handle this error.

    If the backend module exists, but there is not active implementation then a
    BackendNotImplementedError will be raised, it is also the responsibility of
    the caller to handle this error.

    """
    try:
        inst = BACKENDS[backend].backend
    except KeyError:
        raise BackendError('Unknown backend: {}'.format(backend))

    if inst is None:
        raise BackendNotImplementedError(
            'Backend for {} is not implemented'.format(backend))

    return inst


def get_extension(file_path):
    """Get the extension name to use when searching for a loader.

    This function correctly handles compression suffixes, as long as they are
    valid.

    """
    def _extension(file_path):
        """Helper function to get the extension string."""
        compression = 'none'
        name, extension = os.path.splitext(file_path)

        # If we hit a compressed suffix, get an additional suffix to test
        # with.
        # i.e: Use .json.gz rather that .gz
        if extension in COMPRESSION_SUFFIXES:
            compression = extension[1:]  # Drop the leading '.'
            # Remove any trailing '.', this fixes a bug where the filename
            # is 'foo.json..xz, or similar
            extension = os.path.splitext(name.rstrip('.'))[1]

        return extension, compression

    if not os.path.isdir(file_path):
        return _extension(file_path)
    else:
        for file_ in os.listdir(file_path):
            if file_.startswith('result') and not file_.endswith('.old'):
                return _extension(file_)

    tests = os.path.join(file_path, 'tests')
    if os.path.exists(tests):
        return _extension(os.listdir(tests)[0])
    else:
        # At this point we have failed to find any sort of backend, just
        # except and die
        raise BackendError("No backend found for any file in {}".format(
            file_path))


def load(file_path):
    """Wrapper for loading runs.

    This function will attempt to determine how to load the file (based on file
    extension), and then pass the file path into the appropriate loader, and
    then return the TestrunResult instance.

    """

    extension, compression = get_extension(file_path)

    for backend in BACKENDS.values():
        if extension in backend.extensions:
            loader = backend.load

            if loader is None:
                raise BackendNotImplementedError(
                    'Loader for {} is not implemented'.format(extension))

            return loader(file_path, compression)

    raise BackendError(
        'No module supports file extension "{}"'.format(extension))


def write(results, file_path):
    """Wrapper for writing runs.

    This function will attempt to determine how to write a TestrunResult
    instance into the file (based on file extension), passing the TestrunResult
    instance into the appropriate writer.

    Returns whether the backend will attempt to use compression when writing
    the results file or not.
    """

    extension, _ = get_extension(file_path)

    for backend in BACKENDS.values():
        if extension in backend.extensions:
            writer = backend.write

            if writer is None:
                raise BackendNotImplementedError(
                    'Writer for {} is not implemented'.format(extension))

            return writer(results, file_path)

    raise BackendError(
        'No module supports file extension "{}"'.format(extension))


def set_meta(backend, result):
    """Wrapper around meta that gets the right meta function."""
    try:
        BACKENDS[backend].meta(result)
    except KeyError:
        raise BackendError('No backend {}'.format(backend))
    except TypeError as e:
        # Since we initialize non-implemented backends as None, and None isn't
        # callable then we'll get a TypeError, and we're looking for NoneType
        # in the message. If we get that we really want a
        # BackendNotImplementedError
        if str(e) == "'NoneType' object is not callable":
            raise BackendNotImplementedError(
                'meta function for {} not implemented.'.format(backend))
        else:
            # Otherwise re-raise the error
            raise



# coding=utf-8
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Module for results generation """

import collections
import copy
import datetime

from framework import status, exceptions, grouptools

__all__ = [
    'TestrunResult',
    'TestResult',
]


class Subtests(collections.abc.MutableMapping):
    """A dict-like object that stores Statuses as values."""
    def __init__(self, dict_=None):
        self.__container = collections.OrderedDict()

        if dict_ is not None:
            self.update(dict_)

    def __setitem__(self, name, value):
        self.__container[name.lower()] = status.status_lookup(value)

    def __getitem__(self, name):
        return self.__container[name.lower()]

    def __delitem__(self, name):
        del self.__container[name.lower()]

    def __iter__(self):
        return iter(self.__container)

    def __len__(self):
        return len(self.__container)

    def __repr__(self):
        return repr(self.__container)

    def to_json(self):
        res = dict(self)
        res['__type__'] = 'Subtests'
        return res

    @classmethod
    def from_dict(cls, dict_):
        if '__type__' in dict_:
            del dict_['__type__']

        res = cls(dict_)

        return res


class StringDescriptor(object):  # pylint: disable=too-few-public-methods
    """A Shared data descriptor class for TestResult.

    This provides a property that can be passed a str or unicode, but always
    returns a unicode object.

    """
    def __init__(self, name, default=''):
        assert isinstance(default, str)
        self.__name = name
        self.__default = default

    def __get__(self, instance, cls):
        return getattr(instance, self.__name, self.__default)

    def __set__(self, instance, value):
        if isinstance(value, bytes):
            setattr(instance, self.__name, value.decode('utf-8', 'replace'))
        elif isinstance(value, str):
            setattr(instance, self.__name, value)
        else:
            raise TypeError('{} attribute must be a unicode or bytes instance, '
                            'but was {}.'.format(self.__name, type(value)))

    def __delete__(self, instance):
        raise NotImplementedError


class TimeAttribute(object):
    """Attribute of TestResult for time.

    This attribute provides a couple of nice helpers. It stores the start and
    end time and provides methods for getting the total and delta of the times.

    """
    __slots__ = ['start', 'end']

    def __init__(self, start=0.0, end=0.0):
        self.start = start
        self.end = end

    @property
    def total(self):
        return self.end - self.start

    @property
    def delta(self):
        return str(datetime.timedelta(seconds=self.total))

    def to_json(self):
        return {
            'start': self.start,
            'end': self.end,
            '__type__': 'TimeAttribute',
        }

    @classmethod
    def from_dict(cls, dict_):
        dict_ = copy.copy(dict_)

        if '__type__' in dict_:
            del dict_['__type__']
        return cls(**dict_)


class TestResult(object):
    """An object representing the result of a single test."""
    __slots__ = ['returncode', '_err', '_out', 'time', 'command', 'traceback',
                 'environment', 'subtests', 'dmesg', '__result', 'images',
                 'exception', 'pid']
    err = StringDescriptor('_err')
    out = StringDescriptor('_out')

    def __init__(self, result=None):
        self.returncode = None
        self.time = TimeAttribute()
        self.command = str()
        self.environment = str()
        self.subtests = Subtests()
        self.dmesg = str()
        self.images = None
        self.traceback = None
        self.exception = None
        self.pid = []
        if result:
            self.result = result
        else:
            self.__result = status.NOTRUN

    @property
    def result(self):
        """Return the result of the test.

        If there are subtests return the "worst" value of those subtests. If
        there are not return the stored value of the test. There is an
        exception to this rule, and that's if the status is crash; since this
        status is set by the framework, and can be generated even when some or
        all unit tests pass.

        """
        if self.subtests and self.__result != status.CRASH:
            return max(self.subtests.values())
        return self.__result

    @property
    def raw_result(self):
        """Get the result of the test without taking subtests into account."""
        return self.__result

    @result.setter
    def result(self, new):
        try:
            self.__result = status.status_lookup(new)
        except exceptions.PiglitInternalError as e:
            raise exceptions.PiglitFatalError(str(e))

    def to_json(self):
        """Return the TestResult as a json serializable object."""
        obj = {
            '__type__': 'TestResult',
            'command': self.command,
            'environment': self.environment,
            'err': self.err,
            'out': self.out,
            'result': self.result,
            'returncode': self.returncode,
            'subtests': self.subtests.to_json(),
            'time': self.time.to_json(),
            'exception': self.exception,
            'traceback': self.traceback,
            'dmesg': self.dmesg,
            'images': self.images,
            'pid': self.pid,
        }
        return obj

    @classmethod
    def from_dict(cls, dict_):
        """Load an already generated result in dictionary form.

        This is used as an alternate constructor which converts an existing
        dictionary into a TestResult object. It converts a key 'result' into a
        status.Status object

        """
        # pylint will say that assining to inst.out or inst.err is a non-slot
        # because self.err and self.out are descriptors, methods that act like
        # variables. Just silence pylint
        # pylint: disable=assigning-non-slot
        inst = cls()

        for each in ['returncode', 'command', 'exception', 'environment',
                     'traceback', 'dmesg', 'images', 'pid', 'result']:
            if each in dict_:
                setattr(inst, each, dict_[each])

        # Set special instances
        if 'subtests' in dict_:
            inst.subtests = Subtests.from_dict(dict_['subtests'])
        if 'time' in dict_:
            inst.time = TimeAttribute.from_dict(dict_['time'])

        # out and err must be set manually to avoid replacing the setter
        if 'out' in dict_:
            inst.out = dict_['out']
        if 'err' in dict_:
            inst.err = dict_['err']

        return inst

    def update(self, dict_):
        """Update the results and subtests fields from a piglit test.

        Native piglit tests output their data as valid json, and piglit uses
        the json module to parse this data. This method consumes that raw
        dictionary data and updates itself.

        """
        if 'result' in dict_:
            self.result = dict_['result']
            if 'images' in dict_:
                self.images = dict_['images']
        elif 'subtest' in dict_:
            self.subtests.update(dict_['subtest'])


class Totals(dict):
    def __init__(self, *args, **kwargs):
        super(Totals, self).__init__(*args, **kwargs)
        for each in status.ALL:
            each = str(each)
            if each not in self:
                self[each] = 0

    def __bool__(self):
        # Since totals are prepopulated, calling 'if not <Totals instance>'
        # will always result in True, this will cause it to return True only if
        # one of the values is not zero
        for each in self.values():
            if each != 0:
                return True
        return False

    def to_json(self):
        """Convert totals to a json object."""
        result = copy.copy(self)
        result['__type__'] = 'Totals'
        return result

    @classmethod
    def from_dict(cls, dict_):
        """Convert a dictionary into a Totals object."""
        tots = cls(dict_)
        if '__type__' in tots:
            del tots['__type__']
        return tots


class TestrunResult(object):
    """The result of a single piglit run."""
    def __init__(self):
        self.info = {}
        self.options = {}
        self.time_elapsed = TimeAttribute()
        self.tests = collections.OrderedDict()
        self.totals = collections.defaultdict(Totals)

    def get_result(self, key):
        """Get the result of a test or subtest.

        If neither a test nor a subtest instance exist, then raise the original
        KeyError generated from looking up <key> in the tests attribute. It is
        the job of the caller to handle this error.

        Arguments:
        key -- the key name of the test to return

        """
        try:
            return self.tests[key].result
        except KeyError as e:
            name, test = grouptools.splitname(key)
            try:
                return self.tests[name].subtests[test]
            except KeyError:
                raise e

    def calculate_group_totals(self):
        """Calculate the number of passes, fails, etc at each level."""
        for name, result in self.tests.items():
            # If there are subtests treat the test as if it is a group instead
            # of a test.
            if result.subtests:
                for res in result.subtests.values():
                    res = str(res)
                    temp = name

                    self.totals[temp][res] += 1
                    while temp:
                        temp = grouptools.groupname(temp)
                        self.totals[temp][res] += 1
                    self.totals['root'][res] += 1
            else:
                res = str(result.result)
                while name:
                    name = grouptools.groupname(name)
                    self.totals[name][res] += 1
                self.totals['root'][res] += 1

    def to_json(self):
        if not self.totals:
            self.calculate_group_totals()
        rep = copy.copy(self.__dict__)
        rep['tests'] = collections.OrderedDict((k, t.to_json())
                       for k, t in self.tests.items())
        rep['__type__'] = 'TestrunResult'
        return rep

    @classmethod
    def from_dict(cls, dict_, _no_totals=False):
        """Convert a dictionary into a TestrunResult.

        This method is meant to be used for loading results from json or
        similar formats

        _no_totals is not meant to be used externally, it allows us to control
        the generation of totals when loading old results formats.

        """
        res = cls()
        for name in ['name', 'info', 'options', 'results_version']:
            value = dict_.get(name)
            if value:
                setattr(res, name, value)

        # Since this is used to load partial metadata when writing final test
        # results there is no guarantee that this will have a "time_elapsed"
        # key
        if 'time_elapsed' in dict_:
            setattr(res, 'time_elapsed',
                    TimeAttribute.from_dict(dict_['time_elapsed']))
        res.tests = collections.OrderedDict((n, TestResult.from_dict(t))
                    for n, t in dict_['tests'].items())

        if not 'totals' in dict_ and not _no_totals:
            res.calculate_group_totals()
        else:
            res.totals = {n: Totals.from_dict(t) for n, t in
                          dict_['totals'].items()}

        return res



# coding=utf-8
# Copyright (C) 2012, 2014-2016, 2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" This module enables running shader tests. """

import io
import os
import re

from framework import exceptions
from framework import status
from framework import options
from .base import ReducedProcessMixin, TestIsSkip
from .opengl import FastSkipMixin, FastSkip
from .piglit_test import PiglitBaseTest, ROOT_DIR

__all__ = [
    'ShaderTest',
]


class Parser(object):
    """An object responsible for parsing a shader_test file."""

    _is_gl = re.compile(r'GL (<|<=|=|>=|>) \d\.\d')
    _match_gl_version = re.compile(
        r'^GL\s+(?P<profile>(ES|CORE|COMPAT))?\s*(?P<op>(<|<=|=|>=|>))\s*(?P<ver>\d\.\d)')
    _match_glsl_version = re.compile(
        r'^GLSL\s+(?P<es>ES)?\s*(?P<op>(<|<=|=|>=|>))\s*(?P<ver>\d\.\d+)')

    def __init__(self, filename):
        self.filename = filename
        self.extensions = set()
        self.api_version = 0.0
        self.shader_version = 0.0
        self.api = None
        self.prog = None
        self.__op = None
        self.__sl_op = None

    def parse(self):
        # Iterate over the lines in shader file looking for the config section.
        # By using a generator this can be split into two for loops at minimal
        # cost. The first one looks for the start of the config block or raises
        # an exception. The second looks for the GL version or raises an
        # exception
        with io.open(os.path.join(ROOT_DIR, self.filename), 'r', encoding='utf-8') as shader_file:
            lines = (l for l in shader_file.readlines())

            # Find the config section
            for line in lines:
                # We need to find the first line of the configuration file, as
                # soon as we do then we can move on to getting the
                # configuration. The first line needs to be parsed by the next
                # block.
                if line.startswith('[require]'):
                    break
            else:
                raise exceptions.PiglitFatalError(
                    "In file {}: Config block not found".format(self.filename))

        for line in lines:
            if line.startswith('GL_'):
                line = line.strip()
                if not (line.startswith('GL_MAX') or line.startswith('GL_NUM')):
                    self.extensions.add(line)
                    if line == 'GL_ARB_compatibility':
                        assert self.api is None or self.api == 'compat'
                        self.api = 'compat'
                continue

            # Find any GLES requirements.
            if not self.api_version:
                m = self._match_gl_version.match(line)
                if m:
                    self.__op = m.group('op')
                    self.api_version = float(m.group('ver'))
                    if m.group('profile') == 'ES':
                        assert self.api is None or self.api == 'gles2'
                        self.api = 'gles2'
                    elif m.group('profile') == 'COMPAT':
                        assert self.api is None or self.api == 'compat'
                        self.api = 'compat'
                    elif self.api_version >= 3.1:
                        assert self.api is None or self.api == 'core'
                        self.api = 'core'
                    continue

            if not self.shader_version:
                # Find any GLSL requirements
                m = self._match_glsl_version.match(line)
                if m:
                    self.__sl_op = m.group('op')
                    self.shader_version = float(m.group('ver'))
                    if m.group('es'):
                        assert self.api is None or self.api == 'gles2'
                        self.api = 'gles2'
                    continue

            if line.startswith('['):
                if not self.api:
                    # Because this is inferred rather than explicitly declared
                    # check this after al other requirements are parsed. It's
                    # possible that a test can declare glsl >= 1.30 and GL >=
                    # 4.0
                    if self.shader_version < 1.4:
                        self.api = 'compat'
                    else:
                        self.api = 'core'
                break

        # Select the correct binary to run the test, but be as conservative as
        # possible by always selecting the lowest version that meets the
        # criteria.
        if self.api == 'gles2':
            if self.__op in ['<', '<='] or (
                    self.__op in ['=', '>='] and self.api_version is not None
                    and self.api_version < 3):
                self.prog = 'shader_runner_gles2'
            else:
                self.prog = 'shader_runner_gles3'
        else:
            self.prog = 'shader_runner'


class ShaderTest(FastSkipMixin, PiglitBaseTest):
    """ Parse a shader test file and return a PiglitTest instance

    This function parses a shader test to determine if it's a GL, GLES2 or
    GLES3 test, and then returns a PiglitTest setup properly.

    """

    def __init__(self, command, api=None, extensions=set(),
                  shader_version=None, api_version=None, env=None, **kwargs):
        super(ShaderTest, self).__init__(
            command,
            run_concurrent=True,
            api=api,
            extensions=extensions,
            shader_version=shader_version,
            api_version=api_version,
            env=env)

    @classmethod
    def new(cls, filename, installed_name=None):
        """Parse an XML file and create a new instance.

        :param str filename: The name of the file to parse
        :param str installed_name: The relative path to the file when installed
            if not the same as the parsed name
        """
        parser = Parser(filename)
        parser.parse()

        return cls(
            [parser.prog, installed_name or filename],
            run_concurrent=True,
            api=parser.api,
            extensions=parser.extensions,
            shader_version=parser.shader_version,
            api_version=parser.api_version)

    @PiglitBaseTest.command.getter
    def command(self):
        """ Add -auto, -fbo and -glsl (if needed) to the test command """

        command = super(ShaderTest, self).command
        shaderfile = os.path.join(ROOT_DIR, self._command[1])

        if options.OPTIONS.force_glsl:
            return self.keys() + [command[0]] + [shaderfile, '-auto', '-fbo', '-glsl']
        else:
            return self.keys() + [command[0]] + [shaderfile, '-auto', '-fbo']

    @command.setter
    def command(self, new):
        self._command = [n for n in new if n not in ['-auto', '-fbo']]


class MultiShaderTest(ReducedProcessMixin, PiglitBaseTest):
    """A Shader class that can run more than one test at a time.

    This class can call shader_runner with multiple shader_files at a time, and
    interpret the results, as well as handle pre-mature exit through crashes or
    from breaking import assupmtions in the utils about skipping.

    Arguments:
    filenames -- a list of absolute paths to shader test files
    """

    def __init__(self, prog, files, subtests, skips, env=None):
        super(MultiShaderTest, self).__init__(
            [prog] + files,
            subtests=subtests,
            run_concurrent=True,
            env=env)

        self.prog = prog
        self.files = files
        self.subtests = subtests
        self.skips = [FastSkip(**s) for s in skips]

    @classmethod
    def new(cls, filenames, installednames=None):
        # TODO
        assert filenames
        prog = None
        subtests = []
        skips = []

        # Walk each subtest, and either add it to the list of tests to run, or
        # determine it is skip, and set the result of that test in the subtests
        # dictionary to skip without adding it to the list of tests to run.
        for each in filenames:
            parser = Parser(each)
            parser.parse()
            subtests.append(os.path.basename(os.path.splitext(each)[0]).lower())

            if prog is not None:
                # This allows mixing GLES2 and GLES3 shader test files
                # together. Since GLES2 profiles can be promoted to GLES3, this
                # is fine.
                if parser.prog != prog:
                    # Pylint can't figure out that prog is not None.
                    if 'gles' in parser.prog and 'gles' in prog:  # pylint: disable=unsupported-membership-test
                        prog = max(parser.prog, prog)
                    else:
                        # The only way we can get here is if one is GLES and
                        # one is not, since there is only one desktop runner
                        # thus it will never fail the is parser.prog != prog
                        # check
                        raise exceptions.PiglitInternalError(
                            'GLES and GL shaders in the same command!\n'
                            'Cannot pick a shader_runner binary!\n'
                            'in file: {}'.format(os.path.dirname(each)))
            else:
                prog = parser.prog

            skips.append({
                'extensions': parser.extensions,
                'api_version': parser.api_version,
                'shader_version': parser.shader_version,
                'api': parser.api,
            })

        return cls(prog, installednames or filenames, subtests, skips)

    def _process_skips(self):
        r_files = []
        r_subtests = []
        r_skips = []
        for f, s, k in zip(self.files, self.subtests, self.skips):
            try:
                k.test()
            except TestIsSkip:
                r_skips.append(s)
            else:
                r_files.append(f)
                r_subtests.append(s)

        assert len(r_subtests) + len(r_skips) == len(self.files), \
            'not all tests accounted for'

        for name in r_skips:
            self.result.subtests[name] = status.SKIP

        self._expected = r_subtests
        self._command = [self._command[0]] + r_files

    def run(self):
        self._process_skips()
        super(MultiShaderTest, self).run()

    @PiglitBaseTest.command.getter
    def command(self):
        command = super(MultiShaderTest, self).command
        shaderfiles = (x for x in command[1:] if not x.startswith('-'))
        shaderfiles = [os.path.join(ROOT_DIR, s) for s in shaderfiles]
        return self.keys() + [command[0]] + shaderfiles + ['-auto', '-report-subtests']

    def _is_subtest(self, line):
        return line.startswith('PIGLIT TEST:')

    def _resume(self, current):
        command = [self.command[0]]
        command.extend(self.command[current + 1:])
        return command

    def _stop_status(self):
        # If the lower level framework skips then return a status for that
        # subtest as skip, and resume.
        if self.result.out.endswith('PIGLIT: {"result": "skip" }\n'):
            return status.SKIP
        if self.result.returncode > 0:
            return status.FAIL
        return status.CRASH

    def _is_cherry(self):
        # Due to the way that piglt is architected if a particular feature
        # isn't supported it causes the test to exit with status 0. There is no
        # straightforward way to fix this, so we work around it by looking for
        # the message that feature provides and marking the test as not
        # "cherry" when it is found at the *end* of stdout. (We don't want to
        # match other places or we'll end up in an infinite loop)
        return (
            self.result.returncode == 0 and not
            self.result.out.endswith(
                'not supported on this implementation\n') and not
            self.result.out.endswith(
                'PIGLIT: {"result": "skip" }\n'))



#!/usr/local/bin/python3.13
# -*- coding: utf-8 -*-
"""All Vulkan tests that come with piglit, using default settings."""

import os

from framework.profile import TestProfile
from framework import grouptools
from framework.test.piglit_test import VkRunnerTest
from .py_modules.constants import TESTS_DIR, GENERATED_TESTS_DIR

__all__ = ['profile']

profile = TestProfile()

# Find and add all shader tests.
for basedir in [TESTS_DIR, GENERATED_TESTS_DIR]:
    _basedir = os.path.join(basedir, 'vulkan')
    for dirpath, _, filenames in os.walk(_basedir):
        groupname = grouptools.from_path(os.path.relpath(dirpath, _basedir))
        groupname = grouptools.join('vulkan', groupname)
        dirname = os.path.relpath(dirpath, os.path.join(basedir, '..'))
        for filename in filenames:
            testname, ext = os.path.splitext(filename)
            if ext != '.vk_shader_test':
                continue
            test = VkRunnerTest(os.path.join(dirname, filename))
            group = grouptools.join(groupname, testname)
            assert group not in profile.test_list, group

            profile.test_list[group] = test



# coding=utf-8
#
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import argparse

from framework import exceptions
from framework.programs import parsers as piglit_parsers
from framework.replay import backends, options

from . import parsers

__all__ = ['dump']


def _dump_from_trace(args):
    options.OPTIONS.device_name = args.device_name
    options.OPTIONS.results_path = args.output

    if args.calls is not None:
        calls = args.calls.split(",")
    else:
        calls = []

    return backends.dump(args.file_path, args.output, calls)


@exceptions.handler
def dump(input_):
    """ Parser for replayer dump command """
    unparsed = piglit_parsers.parse_config(input_)[1]

    # Set the parent of the config to add the -f/--config message
    parser = argparse.ArgumentParser(parents=[piglit_parsers.CONFIG,
                                              parsers.DEVICE,
                                              parsers.RESULTS_PATH])
    parser.add_argument(
        '-c', '--calls',
        dest='calls',
        required=False,
        default=None,
        help=('a comma separated list of call numbers from the trace to dump '
              '(default: last frame).'))
    parser.add_argument(
        'file_path',
        help=('the path to the local trace file '
              'from which to dump images.'))
    parser.set_defaults(func=_dump_from_trace)

    args = parser.parse_args(unparsed)

    return 0 if args.func(args) else 1
